// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// graa.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_graa_h
#define included_graa_h

#ifndef included_grap_h
#  include <nbdy/grap.h>
#endif

#ifndef included_soul_h
#  include <nbdy/soul.h>
#endif

using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // class nbdy::grape_arrays                                                   
  //                                                                            
  // allocates and holds arrays for the souls' positions, masses, accelerations,
  // potentials, and booleans indicating sink. These are used by the GRAPE      
  // routines utilizing the GRAPE hardware for computation of gravity.          
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class grape_arrays {
    //--------------------------------------------------------------------------
    // member types of grape_arrays                                             
    //--------------------------------------------------------------------------
  public:
    typedef GRAPE_DATA_TYPE         grape_scalar_type;
    typedef grape_scalar_type       grape_array_type[3];
    typedef sym1<grape_scalar_type> grape_vector_type;
    //--------------------------------------------------------------------------
    // data of grape_arrays                                                     
    //--------------------------------------------------------------------------
  private:
    unsigned           N;                            // # souls                 
    const soul        *S0;                           // first soul              
    grape_array_type  *POS;                          // positions               
    grape_scalar_type *MASS;                         // masses                  
    bool              *SINK;                         // flag: need force?       
    grape_array_type  *ACC;                          // accelerations           
    grape_scalar_type *POT;                          // masses                  
    //--------------------------------------------------------------------------
    // const data access                                                        
    //--------------------------------------------------------------------------
  public:
    const grape_vector_type pos(const int&i) const {
      return grape_vector_type(POS[i]); }
    const grape_vector_type acc(const int&i) const {
      return grape_vector_type(ACC[i]); }
    const grape_scalar_type&pot(const int&i) const { return POT[i]; }
    //--------------------------------------------------------------------------
    // construction & destruction                                               
    //--------------------------------------------------------------------------
    grape_arrays(const uint& n,
		 const soul* s0) :
      N    ( n ),
      S0   ( s0 ),
      POS  ( new grape_array_type[N] ),
      MASS ( new grape_scalar_type[N] ),
      SINK ( new bool[N] ),
      ACC  ( new grape_array_type[N] ),
      POT  ( new grape_scalar_type[N] )
    {
      register uint        i;
      register const soul *s;
      for(i=0, s=S0; i!=N; ++i,++s) {
	POS[i][0]  = cofm(s)[0];
	POS[i][1]  = cofm(s)[1];
#if NDIM==3
	POS[i][2]  = cofm(s)[2];
#else
	POS[i][2]  = 0.;
#endif
	MASS[i]    = mass(s);
	SINK[i]    = is_sink(s);
	ACC[i][0]  = 0.;
	ACC[i][1]  = 0.;
	ACC[i][2]  = 0.;
	POT[i]     = 0.;
      }
    }
    //--------------------------------------------------------------------------
    ~grape_arrays() {
      delete[] POS;
      delete[] MASS;
      delete[] SINK;
      delete[] ACC;
      delete[] POT;
    }
    //--------------------------------------------------------------------------
    // access to arrays at certain soul                                         
    //--------------------------------------------------------------------------
    int index(const soul*s) const { return int(s-S0); }
    grape_array_type*  pos_arr (const int&i) const { return POS+i; }
    grape_scalar_type* mass_arr(const int&i) const { return MASS+i; }
    bool*              sink_arr(const int&i) const { return SINK+i; }
    grape_array_type*  acc_arr (const int&i) const { return ACC+i; }
    grape_scalar_type* pot_arr (const int&i) const { return POT+i; }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_graa_h
