// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// rand.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 1994-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// class RandomNumberGenerator base class for random number generators         |
// class Random3         pseudo-random number generator (Press et al.)         |
// class Sobol           quasi-random number generator (Press et al.)          |
//                                                                             |
// class RandomDeviate   base class for a random distribution                  |
// class Uniform         P(x) = 1/|b-a| for x in (a,b)                         |
// class Gaussian        P(x) = Exp[-x^2/(2 sigma^2)]; x in [-oo,oo]           |
// class Exponential     P(x) = Exp[-x]; x in [0,oo)                           |
// class ExpDisk         P(x) = x Exp[-x]; x in [0,oo)                         |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_rand_h
#define included_rand_h

#ifndef included_inln_h
#  include <util/inln.h>
#endif

namespace util {
  //////////////////////////////////////////////////////////////////////////////
  // class util::RandomNumberGenerator                                          
  //////////////////////////////////////////////////////////////////////////////
  class RandomNumberGenerator {
  public:
    virtual double RandomDouble() = 0;
    void   RandomDouble(double& x) { x = RandomDouble(); }
    double operator()  ()          { return RandomDouble(); }
    void   operator()  (double& x) { RandomDouble(x); }
  }; 
  //////////////////////////////////////////////////////////////////////////////
  // class util::RandomDeviate                                                  
  //////////////////////////////////////////////////////////////////////////////
  class RandomDeviate {
  public:
    virtual double operator() () = 0;
    virtual double value      (const double) const = 0;
  }; 
  //////////////////////////////////////////////////////////////////////////////
  // class util::Random3                                                        
  //////////////////////////////////////////////////////////////////////////////
  class Random3 : public RandomNumberGenerator {
    // pseudo-random number generator
  private:
    int   *inext, *inextp;
    long  *ma;
  public:
    Random3(const long);
    virtual ~Random3();
    virtual double RandomDouble();
  };
  //////////////////////////////////////////////////////////////////////////////
  // class util::Sobol                                                          
  //////////////////////////////////////////////////////////////////////////////
  class Sobol : public RandomNumberGenerator {
    // quasi-random number generator
  private:
    int           actl, in;
    unsigned long bits, ix, *v;
    double        fac;
    void error(const char*) const;
    void warning(const char*) const;
  public:
    Sobol(const int=-1, const int=0);
    virtual ~Sobol();
    virtual double RandomDouble();
    void   Reset       ()   { ix = in = 0; }
    int    actual      ()   { return actl; }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class util::Uniform                                                        
  //   fully inline                                                             
  //////////////////////////////////////////////////////////////////////////////
  class Uniform : public RandomDeviate {
    // gives x uniformly in [a,b]
  private:
    RandomNumberGenerator &r;
    double   a,b,ba;
  public:
    Uniform(RandomNumberGenerator* R,                 // random number generator
	    const double A=0., const double B=1.)     // a,b                    
      : r(*R), a(min(A,B)), b(max(A,B)), ba(b-a) {}
    double lower_bound() { return a; }
    double upper_bound() { return b; }
    double operator() () { return a+ba*r(); }
    double value(const double x) const { return (a<=x && x<=b)? 1. : 0.; }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class util::Gaussian                                                       
  //////////////////////////////////////////////////////////////////////////////
  class Gaussian : public RandomDeviate {
    // gives x in [-oo,oo] with probability proportional to exp[-x^2/(2sigma^2)]
  private:
    int      iset;
    double   sig, norm, gset;
    RandomNumberGenerator *R1, *R2;
  public:
    Gaussian(RandomNumberGenerator*,                  // 1st RNG                
	     RandomNumberGenerator*,                  // 2nd RNG                
	     const double=1.);                        // sigma                  
    double operator() ();
    double sigma   () { return sig; }
    double value(const double) const;
  };
  //////////////////////////////////////////////////////////////////////////////
  // class util::Exponential                                                    
  //////////////////////////////////////////////////////////////////////////////
  class Exponential : public RandomDeviate {
    // gives x in [0,oo] with probability proportional to Exp[-x/alpha]         
  private:
    double alf;
    RandomNumberGenerator *Rn;
  public:
    Exponential(RandomNumberGenerator* R,             // RNG                    
                const double a=1.)                    // alpha                  
      : alf(a), Rn(R) {}
    double operator() ();
    double value(const double) const;
  };
  //////////////////////////////////////////////////////////////////////////////
  // class util::ExpDisk                                                        
  //////////////////////////////////////////////////////////////////////////////
  class ExpDisk: public RandomDeviate {
    // gives x in [0,oo] with probability proportional to  x*Exp[-x/h]
  private:
    const int N,N1;
    RandomNumberGenerator *R;
    double    h, hi, hqi, *Y, *P;
    double    ranvar() const;
  public:
    ExpDisk(RandomNumberGenerator*,                   // RNG                    
	    const double=1.);                         // h                      
    virtual ~ExpDisk();
    double value      (const double) const;
    double operator() ()                { return ranvar(); }
    void   operator() (double& x) const { x = ranvar(); }
  };
  //////////////////////////////////////////////////////////////////////////////
}                                                        // namespace util      
//-----------------------------------------------------------------------------+
#endif                                                   // included_inln_h     
