//-----------------------------------------------------------------------------+
//                                                                             |
// yanc.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                        YET ANOTHER N-BODY CODE                              |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <yanc.h>
#include <string>
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <ctime>

#ifdef unix
#  define GIVE_HOST_INFO 1
  extern "C" {
#  include <unistd.h>
#  include <pwd.h>
  }
#endif

#include <nbdy.h>
#include <nbdy/nmio.h>
#include <util/ionl.h>

using std::string;
using std::ios;
using std::endl;
using std::cerr;
using std::exit;
using std::strcmp;
using std::istream;
using std::ostream;
using std::ifstream;
using std::ofstream;
using std::setw;
using std::setfill;
using std::setprecision;

using namespace util;
using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
namespace nbdy{
  //===========================================================================+
  //                                                                           |
  // auxiliary constants and functions for the installation of class ndata     |
  //                                                                           |
  //============================================================================
  inline basic_nbody::soft_type soften(const indx SOFTEN) {
    switch(SOFTEN) {
    case 1:  return basic_nbody::individual_fixed;
    case 2:  return basic_nbody::global_fixed;
    default: return basic_nbody::global_fixed; 
    }
  }
  //----------------------------------------------------------------------------
  const char* opts[] = {"start_data",      //  0
			"theta",           //  1
			"Nreuse",          //  2
			"Ncrit",           //  3
			"root_center",     //  4
			"reuse",           //  5
			"softening",       //  6
			"Nsoft",           //  7
			"eps",             //  8
			"eps_given",       //  9
			"kernel",          // 10
			"scheme",          // 11
			"Nref",            // 12
			"facc",            // 13
			"fpot",            // 14
			"fcom",            // 15
			"time",            // 16
			"hmin",            // 17
			"Nsteps",          // 18
			"give_pot",        // 19
			"give_rho",        // 20
			"data_format",     // 21
			"N",               // 22
			"##","#",          // 23, 24
			"pot_file",        // 25
			"r_max",           // 26
			"data_file"};      // 27
  const unsigned nopt=28;
  //----------------------------------------------------------------------------
  inline void yancwarning(const char*msgs) {
    cerr<<"YANC: WARNING: "<<msgs<<"\n";
  }
  //----------------------------------------------------------------------------
  inline void yancerror  (const char*msgs) {
    cerr<<"YANC: ERROR: "<<msgs<<"\n";
    exit(1);
  }
  //----------------------------------------------------------------------------
  inline nbdy::uint check_eps_range(const real EPS,
				    nbdy::bodies *BODIES)
  {
    register nbdy::uint alright=0u;
    LoopBodies(bodies,BODIES,Bi)
      if(nbdy::eps(Bi) > EPS) {                       // eps out of range ?     
	Bi.eps() = EPS;                               // put it in range        
	alright++;
      }
    return alright;
  }
  //===========================================================================+
  //                                                                           |
  // class nbdy::ndata                                                         |
  //                                                                           |
  // auxiliary for the installation of class yanc                              |
  //                                                                           |
  //============================================================================
  class ndata {
#ifdef ALLOW_NEMO
    int      ND;                             // # NEMO devices                  
    nemo_out*NEMO;                           // NEMO output devices             
  public:
    real     CPU_I;                          // initial CPU time                
#endif
  public:
    string   FILE;                           // input file                      
    string   POTFILE;                        // file for external potential     
    real     THETA;                          // opening angle                   
    uint     NREUSE;                         // max # subsequent falcON::reuse()
    uint     NCRIT;                          // max # bodies in unsplit cell    
    vect     OR, *POR;                       // center of root                  
    uint     NSOFT;                          // # bodies in softening spheres   
    uint     NREF;                           // # bodies for density estimate   
    real     EPS;                            // eps / eps_max                   
    indx     SOFTEN;                         // 0/1/2 for globally fixed,       
                                             // indiv fixed, and indiv adaptive 
    indx     KERN;                           // (0/1/2/3/9) soft kernel         
    real     FACC;                           // factor f_acc for time-stepping  
    real     FPOT;                           // factor f_pot for time-stepping  
    real     FCOM;                           // factor f_com for time-stepping  
    real     TINI;                           // initial simulation time         
    real     TIME;                           // simulation time                 
    bool     EPS_GIVEN;                      // eps_i are given on input        
    io       WRITE;                          // what to write to output         
    int      HMIN;                           // tau_min = 1/2^h0                
    double   RMAX;                           // r_max for gcsim                 
    indx     NSTEPS;                         // number of tau                   
    indx     FORMAT;                         // data format: 0/1 : ascii/binary 
    bodies  *BODIES;                         // bodies themselves               
  private:
    ndata();                                 // not implemented                 
    ndata(const ndata&);                     // not implemented                 
    //--------------------------------------------------------------------------
  private:
    void read_yanc        ()                     // read input using yanc format
    {
      ifstream in;
      if(!open(in,FILE.c_str())) exit(1);
      char option[100];
      register int  tmp_int;
      register uint iopt = 0;
      register uint Ns,Nbody;
      do {
	if(in)
	  skip_char(in,'#') >> option;
	else {
	  cerr<<"# YANC: end of input seen before \"start_data\"\n";
	  return;
	}
	for(iopt=0; iopt<nopt && strcmp(option,opts[iopt]); iopt++);
	switch(iopt) {
	case  0:                         SwallowRestofLine(in); break;
	case  1: in >> THETA;            SwallowRestofLine(in); break;
	case  2: in >> NREUSE;           SwallowRestofLine(in); break;
	case  3: in >> NCRIT;            SwallowRestofLine(in); break;
	case  4: in >> OR; POR = &OR;    SwallowRestofLine(in); break;
	case  5: cerr<<"# YANC: parameter \"reuse\" currently not supported\n";
	                                 SwallowRestofLine(in); break;
	case  6: in >> SOFTEN;           SwallowRestofLine(in); break;
	case  7: in >> NSOFT;            SwallowRestofLine(in); break;
	case  8: in >> EPS;              SwallowRestofLine(in); break;
	case  9: EPS_GIVEN = true;       SwallowRestofLine(in); break;
	case 10: in >> KERN;             SwallowRestofLine(in); break;
	case 11: cerr<<"# YANC: parameter \"scheme\" currently not supported\n";
	                                 SwallowRestofLine(in); break;
	case 12: in >> NREF;             SwallowRestofLine(in); break;
	case 13: in >> FACC;             SwallowRestofLine(in); break;
	case 14: in >> FPOT;             SwallowRestofLine(in); break;
	case 15: in >> FCOM;             SwallowRestofLine(in); break;
	case 16: in >> TIME;             SwallowRestofLine(in); break;
	case 17: in >> HMIN;             SwallowRestofLine(in); break;
	case 18: in >> Ns;               SwallowRestofLine(in);
	         if(Ns>=1) NSTEPS = Ns;
	         else cerr<<"# YANC: Nsteps>0 required\n";      break;
	case 19: in >> tmp_int;
	         if(tmp_int & 1) WRITE |= io::p;
	         if(tmp_int & 2) WRITE |= io::P;
	                                 SwallowRestofLine(in); break;
	case 20: WRITE |= io::r;         SwallowRestofLine(in); break;
	case 21: in >> FORMAT;           SwallowRestofLine(in); break;
	case 22: in >> Nbody;            SwallowRestofLine(in); break;
	case 23: case 24:                SwallowRestofLine(in); break;
	case 25: in >> POTFILE;          SwallowRestofLine(in); break;
	case 26: in >> RMAX;             SwallowRestofLine(in); break;
	case 27: in >> FILE;             read_yanc();           return;
	default: cerr<<"# YANC: unknown parameter \""<<option<<"\" ignored\n";
	                                 SwallowRestofLine(in); break;
	}
      } while(iopt);
      if(BODIES) BODIES->reset(Nbody,SOFTEN!=0);
      else       BODIES = new bodies(Nbody,SOFTEN!=0);
      if(FORMAT == 0)
	BODIES->read_yanc_ascii (in,EPS_GIVEN? io::mxv | io::e : io::mxv);
      else
	BODIES->read_yanc_binary(in,EPS_GIVEN? io::mxv | io::e : io::mxv);
      if(! in) {
	cerr<<"# YANC: too few bodies in input\n";
	delete BODIES;
	BODIES = 0;
	return;
      }
      if(SOFTEN && EPS_GIVEN) {
	uint a = check_eps_range(EPS,BODIES);
	if(NSOFT>0u && a>0) cerr<<"# YANC: "<<a<<" eps_i > eps_max\n";
      }
    }
    //--------------------------------------------------------------------------
  private:
    void write_yanc_header(                       // write header in yanc format
			   ostream   &out,        // I/O:  stream to write to   
			   const char*file,       // I:    name of file written 
			   const bool eps_i,      // I:    add param eps_given  
			   const char*prog,       // I:    name of calling main 
			   const indx format)     // I:    format type          
      const 
    {
#ifdef GIVE_HOST_INFO
      char host[100]; gethostname(host,100);
      time_t now = ::time(0);
#endif
      out.setf(ios::left);
      out<<setfill(' ');
      out<<"## -------------------------------"
	"-----------------------------------\n"
	 <<"##  file          \""<<file<<"\"\n"
#ifdef GIVE_HOST_INFO
	 <<"##  generated on  "  <<ctime(&now);
      if(prog) out<<"##  with program  \""<<prog<<"\"\n";
      out<<"##  by user       \""<<(getpwuid(geteuid())->pw_name)<<"\"\n"
	 <<"##  on machine    \""<<host<<"\"\n";
#else
      ;
      if(prog) out<<"##  generated with program  \""<<prog<<"\"\n";
#endif
      if(!FILE.empty())
	out<<"##\n"
	   <<"##  original data read from file \""<<FILE<<"\"\n";
      out<<"##  initial time = "<<TINI<<"\n"
	 <<"## -------------------------------"
	"-----------------------------------\n";
      // tree issues
      out<<"# theta          "<<setw(15)<<THETA;
      if(THETA<zero) out<<" constant opening angle\n";
      else           out<<" mass-dependent opening angle; theta_min\n";
      out<<"# Nreuse         "<<setw(15)<<NREUSE
	 <<" max # of subsequend tree reuses\n"
	 <<"# Ncrit          "<<setw(15)<<NCRIT
	 <<" max # of bodies in un-splitted cell\n";
      if(POR) out<<"# root_center    "<<setw(4)<<OR[0]<<" "<<setw(4)<<OR[1]<<" "
		 <<setw(4)<<OR[2]<<"  use fixed root center\n";
      if(!POTFILE.empty()) out<<"# pot_file       "<<setw(15)<<POTFILE
		     <<" using external GalaxyPotential\n";
      if(RMAX) out<<"# r_max          "<<setw(15)<<RMAX;
      // softening issues
      out<<"# softening      "<<setw(15)<<SOFTEN;
      switch(SOFTEN) {
      case 1:
	out<<" individually fixed softening lengths\n";
	break;
      case 2:
	out<<" individually adaptive softening lengths with\n"
	   <<"# Nsoft          "<<setw(15)<<NSOFT
	   <<" # bodies in softening sphere\n"
	   <<"# eps            "<<setw(15)<<EPS
	   <<" maximum eps_i\n";
	break;
      default:
	out<<" globally fixed softening length:\n"
	   <<"# eps            "<<setw(16)<<EPS<<"\n";
      }
      // time-stepping issues
      if(NSTEPS==1) 
	out<<"# Nsteps         1               (ordinary leap-frog)\n"
	   <<"# hmin           "<<setw(15)<<HMIN<<" with step tau "
	   <<tau_min()<<"\n";
      else {
	out<<"# facc           "<<setw(15)<<FACC;
	if(FACC) out<<"tau <= "<<FACC<<" / |acc|\n"; else out<<"\n";
	out<<"# fpot           "<<setw(15)<<FPOT;
	if(FPOT) out<<"tau <= "<<FPOT<<" / |pot|\n"; else out<<"\n";
	out<<"# fcom           "<<setw(15)<<FCOM;
	if(FCOM) out<<"tau <= "<<FCOM<<" * sqrt(-pot) / |acc|\n"; else out<<"\n";
	out<<"# hmin           "<<setw(15)<<HMIN<<" tau_min = 2^"<<(-HMIN)
	   <<" = "<<tau_min()<<"\n";
	out<<"# Nsteps         "<<setw(15)<<NSTEPS<<" tau_max = 2^"
	   <<(NSTEPS-HMIN-1)
	   <<" = "<<tau_max()<<"\n";
      }
      // data issues
      out<<"# time           "<<setw(15)<<TIME<<" simulation time\n"
	 <<"# N              "<<setw(15)<<BODIES->N_body()<<" number of bodies\n"
	 <<"# data_format    "<<setw(15)<<format;
      if(format==0) out<<" bodies in yanc ascii format\n";
      else          out<<" bodies in yanc binary format\n";
      if(format==0) {
	int give_pot = 0;
	if(WRITE & io::p) give_pot += 1;
	if(WRITE & io::P) give_pot += 2;
	if(give_pot)
	  out<<"# give_pot       "<<setw(16)<<give_pot
	     <<"potential given on output\n";
	if(WRITE & io::r)
	  out<<"# give_rho                         density given on output\n";
      }
      if(eps_i)
	out<<"# eps_given                        eps_i given with data below\n";
      out<<"## -----------------------------------"
	"--------------------------------\n"
	 <<"# start_data"<<endl;
    }
    //--------------------------------------------------------------------------
#ifdef ALLOW_NEMO
    //--------------------------------------------------------------------------
  public: void open_nemo(const int  d=0,
			 const char*file=0, 
			 const bool resume=false)
    {
      if(d >= ND) nemo_io::error("nemo device does not exist");
      if(resume) {
	if(file==0 || FILE == string(file))
	  (NEMO+d)->open_to_append(FILE.c_str());
	else {
	  (NEMO+d)->open(file);
	  (NEMO+d)->write_history();
	}
      } else if(file) {
	if(FILE == string(file) && FILE != "-")
	  nemo_io::error("out==in; use option resume instead");
	(NEMO+d)->open(file);
	(NEMO+d)->write_history();
      } else
	nemo_io::warning("cannot open nemo output");
    }
    //--------------------------------------------------------------------------
  public: void close_nemo  (const int d=0)
    {
      if(d < ND) (NEMO+d)->close();
    }
    //--------------------------------------------------------------------------
  public: int nemo_devices() const { return ND; }
    //--------------------------------------------------------------------------
  public: bool nemo_is_open(const int d=0) const
    {
      if(d >= ND) return false;
      return (NEMO+d)->is_open();
    }
    //--------------------------------------------------------------------------
  public:
    void read_nemo(                                // read from NEMO snapshot   
		   const io   want  =io::mxv,      //[I: what to read]          
		   const bool resume=false)        //[I: resume old simul?]     
    {
      if(!BODIES) BODIES = new bodies();           // initialize bodies         
      nemo_in NEMO(FILE.c_str());                  // open nemo input           
      NEMO.read_history();                         // read history              
      register io got;                             // what did we get?          
      if(resume) {                                 // IF (resuming old sim)    >
	do {                                       //   do: read particles     >
	  NEMO.open_set(nemo_io::snap);            //     open nemo snapshot    
	  got = BODIES->read_nemo_particles(NEMO,&TIME,want);// read particles  
	  if(NEMO.is_present(nemo_io::diags)) {    //     IF(diags set)      >  
	    NEMO.open_set(nemo_io::diags);         //       open diags set      
	    if(NEMO.is_present(nemo_io::cputime))  //       IF(cpu time there)  
	      CPU_I=NEMO.read(nemo_io::cputime)*60;//         read CPU [min]    
	    NEMO.close_set(nemo_io::diags);        //       close diags set     
	  }                                        //     <                     
	  NEMO.close_set(nemo_io::snap);           //     close nemo snapshot   
	} while(NEMO.is_present(nemo_io::snap));   //   < while more to be read 
      } else {                                     // < ELSE (not resuming) >   
	  NEMO.open_set(nemo_io::snap);            //   open nemo snapshot      
	  got = BODIES->read_nemo_particles(NEMO,&TIME,want);// read particles  
	  NEMO.close_set(nemo_io::snap);           //   close nemo snapshot     
      }
      if(got & io::e) EPS_GIVEN = true;            // set EPS_GIVEN             
    }
    //--------------------------------------------------------------------------
  public:
    void write_nemo (const io          w=io::mxv,  //[I: what to write out]     
		     const basic_nbody*NBODY=0,    //[I: nbody for diags O]     
		     const int         d    =0)    //[I: index of nemo stream]  
    {
      if(d >= ND)
	nemo_io::error("nemo device does not exist");
      nemo_out *OUT=NEMO+d;
      if(!OUT->is_open())
	nemo_io::error("nemo output stream not opened; cannot write snapshot");
      if(!BODIES) {
	nemo_io::warning("no body exist; cannot write snapshot");
	return;
      }
      register uint        i,j;
      OUT->open_set(nemo_io::snap);                 // open a new nemo snapshot 
      BODIES->write_nemo_particles(*OUT, &TIME,     //   write out particles    
				   SOFTEN? w | io::e : w);
      if(NBODY) {                                   //   IF diags output wantd >
	OUT->open_set(nemo_io::diags);              //     open diagnostics set 
	OUT->single_vec(0)=NBODY->total_energy();   //       copy total energy  
	OUT->single_vec(1)=NBODY->kin_energy();     //       copy kin energy    
	OUT->single_vec(2)=NBODY->pot_energy();     //       copy pot energy    
	OUT->write(nemo_io::energy);                //       write energies     
	for(i=0;i!=NDIM;++i) for(j=0;j!=NDIM;++j)   //     loop dims twice  > > 
	  OUT->single_mat(i,j) = NBODY->kin_energy(i,j);//   copy K_ij      < < 
	OUT->write(nemo_io::KinT);                  //     write K_ij           
	for(i=0;i!=NDIM;++i) for(j=0;j!=NDIM;++j)   //     loop dims twice  > > 
	  OUT->single_mat(i,j) = NBODY->pot_energy(i,j);//   copy W_ij      < < 
	OUT->write(nemo_io::PotT);                  //     write W_ij           
	for(i=0;i!=NDIM;++i) for(j=0;j!=NDIM;++j)   //     loop dimensions  > > 
	  OUT->single_mat(i,j) = NBODY->as_angmom(i,j); //   copy A_ij      < < 
	OUT->write(nemo_io::AmT);                   //     write A_ij           
	for(i=0;i!=NDIM;++i) {                      //     loop dims           <
	  OUT->single_phs(0,i)=NBODY->center_of_mass()[i];// copy c-of-m pos    
	  OUT->single_phs(1,i)=NBODY->total_momentum()[i];// copy c-of-m vel    
	}                                           //     <                    
	OUT->write(nemo_io::cofm);                  //     write c-of-m (x,v)   
	OUT->write(nemo_io::cputime,                //     write accum CPU [min]
		      NBODY->cpu_total()/60.);      //                          
	OUT->close_set(nemo_io::diags);             //     close diagnostics set
      }                                             //   <                      
      OUT->close_set(nemo_io::snap);                // close nemo snapshot      
      OUT->reset();                                 // reset nemo output        
    }
    //--------------------------------------------------------------------------
  public:
    ndata(const char    *file,
	  const bool     nemo,
	  const double   th,
	  const int      nu,
	  const int      nc,
	  const double   ep,
	  const int      kn,
	  const int      hm,
	  const int      ns,
	  const double   fa,
	  const double   fp,
	  const double   fc,
	  const int      nr,
	  const double   rm,
	  const bool     resume  =false,     	    // only applies if nemo=true
	  const bool     read_aux=false,            // only applies if nemo=true
	  const int      nd      =2)                // # nemo devices           
      :
      ND        (nemo? nd : 0),
      NEMO      (new nemo_out[ND]),
      CPU_I     (zero),
      FILE      (file),
      THETA     (real(th)),
      NREUSE    (uint(nu)),
      NCRIT     (uint(nc)),
      POR       (0),
      NSOFT     (0u),
      NREF      (uint(nr)),
      EPS       (real(ep)),
      SOFTEN    (0),
      KERN      (indx(kn)),
      FACC      (real(fa)),
      FPOT      (real(fp)),
      FCOM      (real(fc)),
      TIME      (zero),
      EPS_GIVEN (false),
      WRITE     (io::mxv),
      HMIN      (hm),
      RMAX      (rm),
      NSTEPS    (indx(ns)),
      FORMAT    (0),
      BODIES    (0)
    {
      if(nemo) {
	register io  need  = io::mxv;
	if(read_aux) need |= io::y;
	if(SOFTEN)   need |= io::e;
	read_nemo(need, resume);
      } else
	read_yanc();
      TINI = TIME;
    }
#endif
    //--------------------------------------------------------------------------
  public:
    ndata(const char *file) :
      FILE      (file),
      THETA     (half),
      NREUSE    (0u),
      NCRIT     (6u),
      POR       (0),
      NSOFT     (0u),
      NREF      (16u),
      EPS       (zero),
      SOFTEN    (0),
      KERN      (1),
      FACC      (0.01),
      FPOT      (zero),
      FCOM      (zero),
      TIME      (zero),
      EPS_GIVEN (false),
      WRITE     (io::mxv),
      HMIN      (6),
      RMAX      (0.0),
      NSTEPS    (indx(1)),
      FORMAT    (0),
      BODIES    (0)
    {
      read_yanc();
      TINI = TIME;
    }
    //--------------------------------------------------------------------------
  public:  ~ndata()
    {
#ifdef ALLOW_NEMO
      for(register int d=0; d!=ND; ++d) close_nemo(d);
      delete[] NEMO;
#endif
      if(BODIES) delete BODIES;
    }
    //--------------------------------------------------------------------------
  public:
    void describe(                                 // describe parameter        
		  ostream   & out,                 // I: output stream          
		  const char* name,                // I: name of main           
		  const char* output)              // I: base name: output files
    {
      if(!okay()) {
	out<<"#  ERROR occured during input of parameters and bodies\n";
	return;
      }
      out<<
	"# =======================================================================";
      if(name) out<<"\n# \""<<name<<"\"";
      out<<"\n#  parameters:"
	// tree issues
	 <<"\n#   theta                "<<THETA
	 <<"\n#   Nreuse               "<<NREUSE
	 <<"\n#   Ncrit                "<<NCRIT;
      if(POR) out<<"\n#   center of root       "<<(*POR);
      if(!POTFILE.empty()) out<<"\n#   pot_file            "<<POTFILE;
      if(RMAX)             out<<"\n#   r_max               "<<RMAX;
      // softening issues
      switch(SOFTEN) {
      case 1:
	out<<"\n#   softening            individually fixed";
	if(EPS_GIVEN) out<<"\n#   eps_i                given with data";
	else          out<<"\n#   eps_i                to be initialized using"
			 <<"\n#   Nsoft                "<<NSOFT
			 <<"\n#   eps_max              "<<EPS;
	break;
      case 2:
	out<<"\n#   softening            individually adaptive"
	   <<"\n#   Nsoft                "<<NSOFT
	   <<"\n#   eps_max              "<<EPS;
	break;
      default:
	out<<"\n#   global eps           "<<EPS;
	break;
      }
      out<<"\n#   kernel               "<<falcON::describe(kern(KERN));
      // time-stepping issues
      if(NSTEPS==1)
	out<<"\n#   time step            "<<tau_min();
      else {
	out<<"\n# facc           "<<setw(15)<<FACC;
	if(FACC) out<<"tau <= "<<FACC<<" / |acc|";
	out<<"\n# fpot           "<<setw(15)<<FPOT;
	if(FPOT) out<<"tau <= "<<FPOT<<" / |pot|";
	out<<"\n# fcom           "<<setw(15)<<FCOM;
	if(FCOM) out<<"tau <= "<<FCOM<<" * sqrt(-pot) / |acc|";
	out<<"\n#   tau_min              "<<tau_min()
	   <<"\n#   tau_max              "<<tau_max();
      }
      // data issues
      out<<"\n#   time                 "<<TIME
	 <<"\n#   N                    "<<BODIES->N_body()
	 <<"\n#   input file           "<<FILE;
      if(output)
	out<<"\n#   output files         "<<output<<".XXXX";
      if(FORMAT)  out<<"\n#   data format          yanc binary";
      else        out<<"\n#   data format          yanc ascii";
      if(WRITE & io::p) out<<"\n#   N-body potential     written on output";
      if(WRITE & io::P) out<<"\n#   external potential   written on output";
      if(WRITE & io::r) out<<"\n#   density              written on output";
      out<<endl;
    }
    //--------------------------------------------------------------------------
  public:
    static void describe_yanc_format(ostream& o, const bool GCSIM) {
      o<<
	" Format of YANC I/O files\n"
	" ========================\n"
	" The file starts with a header with one line per parameter.  The syntax\n"
	" of these lines is \"[#] PARAMETER VALUE(S) COMMENT\", where PARAMETER is\n"
	" a character string, see below, while  VALUE is the  numerical value of\n"
	" of the associated parameter. COMMENT is anything between VALUE and EOL\n"
	" and is ignored, as is an optional \"#\" at  the beginning of the  header\n"
	" lines. The meaning of the various parameters is as follows.\n\n"

	" PARAMETER   VALUE default meaning\n"
	" ---------------------------------------------------------------------\n"
	" ##                      this line is (quietly) ignored\n"
	" data_file    FILE       ignore rest of file and use FILE instead\n";
      if(GCSIM)
	o<<
	  " pot_file     FILE       external GalaxyPotential for GCSIM\n"
	  " r_max          rm    0  r_max for bodies in tree\n";
      o<<
	" theta          th  0.5  opening angle  < 0 -> th=|VALUE| = const\n"
	"                                        > 0 -> th=th(M), th_min=TH\n"
	" Nreuse         Nr    0  max # subsequent tree re-uses\n"
	" Ncrit          Nc    6  max # bodies in un-splitted cell\n"
	" root_center X,Y,Z       if specified: center of root cell\n"
	" softening       S    0  determines usage of softening lengths\n"
	"                           0 -> globally fixed\n"
	"                           1 -> individually fixed (NOT TESTED)\n"
	"                           2 -> individually adapted (NOT TESTED)\n"
	" Nsoft          Ns    0  # bodies in softening spheres (for S>0)\n"
	" eps            ep    0  fixed / maximum softening length\n"
	" eps_given               individual eps_i are given on input\n"
	" kernel         kr    1  P_kr (P_0=Plummer) softening kernel\n"
	" Nref           Nf   16  # bodies in density estimate with scheme=1\n"
	" facc           fa 0.01  factor in time-stepping  tau < fa/acc\n"
	" fpot           fp    0  factor in time-stepping  tau < fp/pot\n"
	" fcom           fc    0  factor in time-stepping  tau < fc*sqrt(pot)/acc\n"
	" hmin           hm    6  tau_min = (1/2)^hm\n"
	" Nsteps         Nt    1  # time steps (1 -> leap-frog)\n"
	" give_pot        g    0  0/1/2/3 give no/N-body pot/external pot/both\n"
	" give_rho                give mass density on output\n"
	" time            t    0  initial simulation time\n"
	" N               n    0  number of bodies to be read\n"
	" data_format   for    0  0/1 -> body data are in ascii/binary\n"
	" start_data              this terminates the header\n\n"
	" A parameter already  set may be superseeded  by another line.  A line\n"
	" with unknown PARAMETER is notified but otherwise ignored.\n\n"
	" The body data consist of an integer, whose bits indicate which data\n"
	" are actually given (see below) plus N lines (ASCII/binary) containing\n"
	" these data in this order (if given):\n"
	"   mass         (1    float) :  bit 0,  value   1\n"
	"   position     (NDIM floats):  bit 1,  value   2\n"
	"   velocity     (NDIM floats):  bit 2,  value   4\n"
	"   eps_i        (1    float) :  bit 3,  value   8\n"
	"   potential    (1    float) :  bit 4,  value  16\n"
	"   acceleration (NDIM floats):  bit 5,  value  32\n"
	"   density      (1    float) :  bit 6,  value  64\n"
	"   auxiliary    (1    float) :  bit 7,  value 128\n"
	"   external pot (1    float) :  bit 8,  value 256\n";
    }
    //--------------------------------------------------------------------------
  public:
    bool write_yanc_ascii (                        // write yanc format, ascii  
			   const char* file,       // I:  name: output file     
			   const char* prog) const // I:  name: calling program 
    {
      ofstream out;
      if(!open(out,file)) {
	yancwarning("cannot output");
	return false;
      }
      if(!okay()) {
	yancwarning("not okay(); cannot output");
	return false;
      }
      write_yanc_header(out,file,SOFTEN,prog,0);
      BODIES->write_yanc_ascii(out,SOFTEN? WRITE | io::e : WRITE);
      if(out) return true;
      else    return false;
    }
    //--------------------------------------------------------------------------
  public:
    bool write_yanc_binary(                        // write yanc format, binary 
			   const char* file,       // I:  name: output file     
			   const char* prog) const // I:  name: calling program 
    {
      ofstream out;
      if(!open(out,file)) {
	cerr<<"YANC: write_yanc_binary(): cannot output\n";
	return false;
      }
      if(!okay()) {
	out<<" YANC: not okay(), cannot write_yanc_binary()\n";
	return false;
      }
      write_yanc_header(out,file,SOFTEN,prog,1);
      BODIES->write_yanc_binary(out);
      if(out) return true;
      else    return false;
    }
    //------------------------------------------------------------------------
    // informators
    bool          okay ()       const { return BODIES!=0; }
    const indx&   Nsteps()      const { return NSTEPS; }
    const uint&   Nbodies()     const { return BODIES->N_body(); }
    const uint&   Nref()        const { return NREF; }
    const uint&   Nreuse()      const { return NREUSE; }
    const uint&   Ncrit()       const { return NCRIT; }
    const string& data_file()   const { return FILE; }
    const real&   eps()         const { return EPS; }
    const real    tau_min()     const { return pow(half,HMIN); }
    const real    tau_max()     const { return pow(half,HMIN+1-NSTEPS); }
    const io&     writing()     const { return WRITE; }
    indx          softening()   const { return SOFTEN; }
    indx          kernel()      const { return KERN; }
    const real&   time  ()      const { return TIME; }
    const real&   initial_time()const { return TINI; }
    const real&   theta()       const { return THETA; }
    const real&   facc()        const { return FACC; }
    const real&   fpot()        const { return FPOT; }
    const real&   fcom()        const { return FCOM; }
  };
  }
//=============================================================================+
//                                                                             |
// class nbdy::yanc                                                            |
//                                                                             |
//==============================================================================
inline void check(const ndata* NDATA, const char* m)
{
  if(! NDATA->okay() ) {
    cerr<<"YANC: not okay. cannot "<<m<<"\n";
    exit(1);
  }
}
//------------------------------------------------------------------------------
void yanc::describe(ostream   & out,
		    const char* name,
		    const char* outf)
{
  MYNDATA->describe(out,name,outf);
  out<<"#   total energy:        "<<MYNBODY->total_energy()
     <<
    "\n# ======================================================================="
    "\n#\n";
}
//------------------------------------------------------------------------------
bool yanc::write_yanc_ascii(const char* file, const char* prog)
{
  if(MYNDATA->WRITE & io::r) MYNBODY->estimate_mass_densities();
  return MYNDATA->write_yanc_ascii(file,prog);
}
//------------------------------------------------------------------------------
bool yanc::write_yanc_binary(const char* file, const char* prog) const
{
  return MYNDATA->write_yanc_binary(file,prog);
}
//------------------------------------------------------------------------------
void yanc::describe_yanc_format(ostream& o,const bool gc)
{
  return ndata::describe_yanc_format(o,gc);
}
//------------------------------------------------------------------------------
void yanc::full_step() 
{
  check(MYNDATA, "full_step()");
  MYNBODY->full_step(MYNDATA->POR);
  MYNDATA->TIME = MYNBODY->time();
}
//------------------------------------------------------------------------------
void yanc::long_stats  (ostream&o) const
{
  MYNBODY->long_stats(o);
}
//------------------------------------------------------------------------------
void yanc::short_stats (ostream&o) const
{
  o<<" "; MYNBODY->short_stats(o);
}
//------------------------------------------------------------------------------
void yanc::short_head  (ostream&o) const
{
  o<<"#"; MYNBODY->short_stat_head(o);
  o<<"#"; MYNBODY->short_stat_line(o);
}
//------------------------------------------------------------------------------
#define FUNC(TYPE,NAME) TYPE yanc::NAME() const { return TYPE(MYNDATA->NAME()); }
FUNC(bool, okay)
FUNC(int,  Nsteps)
FUNC(int,  Nbodies)
FUNC(int,  Nref)
FUNC(int,  Nreuse)
FUNC(int,  Ncrit)
FUNC(io,   writing)
FUNC(int,  softening)
FUNC(int,  kernel)
FUNC(float,time)
FUNC(float,initial_time)
FUNC(float,theta)
FUNC(float,facc)
FUNC(float,fpot)
FUNC(float,fcom)
FUNC(float,eps)
FUNC(float,tau_min)
FUNC(float,tau_max)
#undef FUNC
float       yanc::cpu_total() const { return MYNBODY->cpu_total(); }
const char* yanc::data_file() const { return MYNDATA->data_file().c_str(); }
//------------------------------------------------------------------------------
#ifdef ALLOW_NEMO
//------------------------------------------------------------------------------
int yanc::nemo_devices() const {
  return MYNDATA->nemo_devices();
}
//------------------------------------------------------------------------------
bool yanc::nemo_is_open(const int d) const {
  return MYNDATA->nemo_is_open(d);
}
//------------------------------------------------------------------------------
void yanc::open_nemo (const int d, const char *file, const bool resume) {
  MYNDATA->open_nemo(d,file,resume);
}
//------------------------------------------------------------------------------
void yanc::close_nemo(const int d) { MYNDATA->close_nemo(d); }
//------------------------------------------------------------------------------
void yanc::write_nemo(const io w, const int d, const bool di)
{
  if(w & io::r) MYNBODY->estimate_mass_densities();
  MYNDATA->write_nemo(w,di? MYNBODY : 0,d);
}
//------------------------------------------------------------------------------
void yanc::describe_nemo(ostream   & out,
			 const char* com)
{
  if(!okay()) {
    out<<"ERROR something wrong with N-body data, perhaps not read yet?\n";
    exit(1);
  }
  out<<"#"; MYNBODY->short_stat_line(out);
  out<<"# \""<<com<<"\"\n#"<<endl;
}
//------------------------------------------------------------------------------
yanc::yanc(const char     *in,                      // I: data input file       
	   const bool      nemo,                    // I: do nemo I/O?          
	   const double    th,                      // I: theta                 
	   const int       nu,                      // I: Nreuse                
	   const int       nc,                      // I: Ncrit                 
	   const double    ep,                      // I: eps                   
	   const int       kn,                      // I: kern                  
	   const int       hm,                      // I: hmin                  
	   const int       ns,                      // I: Nsteps                
	   const double    fa,                      // I: f_a                   
	   const double    fp,                      //[I: f_p]                  
	   const double    fc,                      //[I: f_c]                  
	   const int       nr,                      //[I: Nref]                 
	   const bool      resume,                  //[I: resume old sim (nemo)]
	   const potential*pt,                      //[I: external potential]   
	   const double    rm,                      //[I: r_max]                
	   const bool      read_aux,                //[I: read aux data?]       
	   const int       nd,                      //[I: # nemo output streams]
	   const uint      dir[4]                   //[I: direct sum control]   
#ifdef __GRAPE
	   ,const bool     gr                       //[I: use GRAPE?]           
#endif
	   ) :
  MYNDATA( new ndata(in,nemo,th,nu,nc,ep,kn,hm,ns,
		     fa,fp,fc,nr,rm,resume,read_aux,nd)),
  MYNBODY( MYNDATA->NSTEPS > 1 ?
	   static_cast<basic_nbody*>(
	     new BlockStepCode(MYNDATA->BODIES,
			       MYNDATA->NSOFT,
			       MYNDATA->EPS,
			       MYNDATA->TIME,
			       MYNDATA->HMIN+1-MYNDATA->NSTEPS,
			       MYNDATA->NSTEPS,
			       MYNDATA->NREUSE,
			       MYNDATA->FACC,
			       MYNDATA->FPOT,
			       MYNDATA->FCOM,
			       MYNDATA->POR,
			       soften(MYNDATA->SOFTEN),
			       MYNDATA->THETA,
			       MYNDATA->NCRIT,
			       kern(MYNDATA->KERN),
			       MYNDATA->NREF,
			       pt,dir
#ifdef __GRAPE
			       ,gr
#endif
			       ) )
	   :
	   static_cast<basic_nbody*>(
	     new LeapFrogCode( MYNDATA->BODIES,
			       MYNDATA->NSOFT,
			       MYNDATA->EPS,
			       MYNDATA->TIME,
			       MYNDATA->HMIN,
			       MYNDATA->NREUSE,
			       MYNDATA->POR,
			       soften(MYNDATA->SOFTEN),
			       MYNDATA->THETA,
			       MYNDATA->NCRIT,
			       kern(MYNDATA->KERN),
			       MYNDATA->NREF,
			       pt,dir
#ifdef __GRAPE
			       ,gr
#endif
			       ) )
	   )
{
  if(resume) MYNBODY->reset_cpu_total(MYNDATA->CPU_I);
}
#endif
//------------------------------------------------------------------------------
yanc::yanc(const char  *in,                         // I: input file            
	   pot_provider*pv) :                       //[I: provider for ext pot] 
  MYNDATA( new ndata(in) ),
  MYNBODY( MYNDATA->NSTEPS > 1 ?
	   static_cast<basic_nbody*>(
	     new BlockStepCode(MYNDATA->BODIES,
			       MYNDATA->NSOFT,
			       MYNDATA->EPS,
			       MYNDATA->TIME,
			       MYNDATA->HMIN+1-MYNDATA->NSTEPS,
			       MYNDATA->NSTEPS,
			       MYNDATA->NREUSE,
			       MYNDATA->FACC,
			       MYNDATA->FPOT,
			       MYNDATA->FCOM,
			       MYNDATA->POR,
			       soften(MYNDATA->SOFTEN),
			       MYNDATA->THETA,
			       MYNDATA->NCRIT,
			       kern(MYNDATA->KERN),
			       MYNDATA->NREF,
			       pv? pv->get_pot(MYNDATA->POTFILE.c_str()) : 0,
			       DIRECT) )
	   :
	   static_cast<basic_nbody*>(
	     new LeapFrogCode (MYNDATA->BODIES,
			       MYNDATA->NSOFT,
			       MYNDATA->EPS,
			       MYNDATA->TIME,
			       MYNDATA->HMIN,
			       MYNDATA->NREUSE,
			       MYNDATA->POR,
			       soften(MYNDATA->SOFTEN),
			       MYNDATA->THETA,
			       MYNDATA->NCRIT,
			       kern(MYNDATA->KERN),
			       MYNDATA->NREF,
			       pv? pv->get_pot(MYNDATA->POTFILE.c_str()) : 0,
			       DIRECT) )
	   ) {}
//------------------------------------------------------------------------------
yanc::~yanc()
{
  delete MYNBODY;
  delete MYNDATA;
}
//------------------------------------------------------------------------------
const bodies* yanc::mybodies() const
{
  return MYNDATA->BODIES;
}
//------------------------------------------------------------------------------
const char* yanc::pot_file() const
{
  return MYNDATA->POTFILE.c_str();
}
//------------------------------------------------------------------------------
const double yanc::r_max() const
{
  return MYNDATA->RMAX;
}
//------------------------------------------------------------------------------
